from typing import List
#  To allow from insel import Insel, InselError in tests:
from .insel import Insel as Insel
from .insel_error import InselError as InselError

from .insel import Parameter
from .existing_model import ExistingModel
from .one_block_model import OneBlockModel
from .template import Template

__version__ = "0.0.8"

# TODO: Add gnuplot functions


def block(name: str, *inputs: float,
          parameters: List[Parameter] = [],
          outputs: int = 1):
    """
    Returns the output of INSEL block *name*, with the given inputs and parameters.
    One output is returned by default, but more can be returned if desired.

    * If the block returns one value, this function returns a float.
    * If the block returns multiple values on one line, this function returns a list of floats.
    * If the block returns multiple lines, this function returns a list of floats.
    * If the block returns multiple values on multiple lines,
         this function returns a list of list of floats.

    >>> insel.block('pi')
    3.141593
    >>> insel.block('sum', 2, 3)
    5.0
    >>> insel.block('do', parameters=[1, 10, 1])
    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
    >>> insel.block('do', parameters=[1, 10, 1])
    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
    >>> insel.block('gain', 2, 5, 7, parameters=[3], outputs=3)
    [6.0, 15.0, 21.0]
    """
    return OneBlockModel(name, inputs=inputs, outputs=outputs, parameters=parameters).run()


def template(template_path, **parameters):
    """
    Returns the output of INSEL template found at template_path,
    after substituting parameters inside the template.

    If template_path is an absolute path, the corresponding template will be used.
    If template_path is a relative path, the template will be searched in templates/ folder.


    >>> insel.template('a_times_b', a=7, b=3)
    21.0
    >>> insel.template('photovoltaic/i_sc',
          pv_id='008823', temperature=25, irradiance=1000)
    5.87388
    """
    return Template(template_path, **parameters).run()


def run(path):
    """Returns the output of INSEL model found at path, without
    substituting any parameter.
    The output is parsed, and returned as float, list of floats or list of list of floats.

    >>> insel.run('templates/one_to_ten.insel')
    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
    """
    return ExistingModel(path).run()


def raw_run(*params):
    """Returns the output of INSEL model found at path, without
    substituting any parameter.
    The output is returned as is, without being parsed.

    >>> print(insel.raw_run('templates/one_to_ten.insel'))
    Compiling one_to_ten.insel ...
    0 errors, 0 warnings
    Running INSEL 8.3.0.9b ...
        1
        2
        3
        4
        5
        6
        7
        8
        9
        10
    Normal end of run
    """
    return ExistingModel(*params).raw_results().decode()



"""
Run all INSEL tests.
Alternative to:
    python -m unittest discover src/
or:
    pytest

This script should work even if insel isn't installed as a package,
while `python test_insel.py` might fail.
"""
import unittest
from pathlib import Path
SRC_DIR = Path(__file__).resolve().parent.parent.parent


def suite():
    loader = unittest.TestLoader()
    return loader.discover(SRC_DIR)


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite(), )
    # COULD DO:
    # print(f'Total INSEL calls : {Insel.calls}')



import platform
import contextlib
import logging
import os
from pathlib import Path
from typing import List
import insel

logging.basicConfig(level=logging.ERROR)

SCRIPT_DIR = Path(__file__).resolve().parent
IS_WINDOWS = platform.system().lower() == 'windows'
STUTTGART = [48.77, 9.18, 1]  # type: List[insel.Parameter]
IMPORTANT_BLOCKS = ['MUL', 'PI', 'PVI', 'MPP', 'DO', 'CLOCK']


@contextlib.contextmanager
def cwd(path):
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)



# coding=utf8
import unittest
import insel


class TestInselFlags(unittest.TestCase):
    def test_insel(self):
        just_insel = insel.raw_run()
        for part in [r'This is INSEL \d\.\d\.\d', '(32|64) bit for (Linux|Windows|macOS)',
                     '-d', '-l', '-m', '-v', '-b']:
            self.assertRegex(just_insel, part,
                             f"'{part}' should be printed out by 'insel'")

    def test_insel_v(self):
        insel_v = insel.raw_run('-v')
        for part in ['libInselEngine', 'libInselBridge', 'libInselTools',
                     'libInselFB', 'libInselEM', 'libInselSE', r'_20\d\d\-\d\d\-\d\d_',
                     # gcc __DATE__ __TIME__ format. e.g. "Mar 31 2022 13:42:25"
                     r'[A-Z][a-z][a-z] [ \d]\d 20\d\d \d\d:\d\d:\d\d']:
            self.assertRegex(insel_v, part,
                             f"'{part}' should be printed out by 'insel -v'")

    def test_insel_l(self):
        insel_l = insel.raw_run('-l', 'templates/one_to_ten.insel')
        for part in [r'1\s*DO\s*T', r'2\s*SCREEN\s*S']:
            self.assertRegex(insel_l, part,
                             f"'{part}' should be printed out by 'insel -l'")

    def test_insel_s(self):
        insel_s = insel.raw_run('-s', 'templates/io/short_string.vseit')
        self.assertRegex(insel_s, '0 errors, 0 warnings',
                         "insel -s should check model")

    def test_insel_m(self):
        insel_m = insel.raw_run('-m', 'templates/io/short_string.vseit')
        for part in [r'b\s+1\s+DO', r'b\s+2\s+SCREEN', "'*'", "'ShortString'"]:
            self.assertRegex(insel_m, part,
                             f"'{part}' should be printed out by 'insel -l'")

    def test_insel_d(self):
        insel_d = insel.raw_run('-d', 'templates/one_to_ten.insel')
        for part in ['Compiling', 'Constructor call', 'Destructor call', 'Standard call',
                     'block DO', 'block SCREEN']:
            self.assertRegex(insel_d, part,
                             f"'{part}' should be printed out by 'insel -d'")



# coding=utf8
from pathlib import Path
import insel
from insel import Insel, InselError
from .custom_assertions import CustomAssertions
from .constants import SCRIPT_DIR, cwd


class TestExistingModel(CustomAssertions):
    def test_one_to_ten(self):
        self.compareLists(
            insel.run('templates/one_to_ten.insel'), range(1, 11))

    def test_screen(self):
        self.compareLists(insel.run('templates/io/screen.insel'), [1, 2, 3])
        header_content = insel.raw_run('templates/io/screen_header.insel')
        self.assertTrue('# HASHTAG' in header_content)
        self.assertTrue('! EXCLAMATION' in header_content)

    def test_screen1g(self):
        self.compareLists(insel.run('templates/io/screen1g.insel'), [])

    def test_add_negative_inputs(self):
        # Little known feature. Could be deleted.
        self.assertEqual(insel.template('add_negative_inputs.insel'), -8)

    def test_nonexisting_model(self):
        self.assertRaisesRegex(InselError, "File not found",
                               insel.run, 'templates/not_here.insel')
        self.assertRaisesRegex(InselError, "File not found",
                               insel.run, 'not_here/model.insel')

    def test_not_an_insel_file(self):
        self.assertRaisesRegex(InselError, "Invalid INSEL model file extension",
                               insel.run, 'data/gengt_comparison.dat')
        self.assertRaisesRegex(InselError, "Invalid INSEL model file extension",
                               insel.run, 'not_even_a_file.csv')

    def test_insel_constants(self):
        self.assertEqual(insel.run('templates/insel_constants.insel'), 3)

    def test_insel_duplicate_constant(self):
        self.assertEqual(
            insel.run('templates/duplicate_constant.insel'), 12345)
        self.assertEqual(Insel.last_warnings,
                         ['W04024 Redefinition of constant TEST skipped'])

    def test_insel_empty_constant(self):
        self.assertEqual(insel.run('templates/empty_constant.insel'), 12345)
        self.assertRegex(Insel.last_raw_output,
                         ("W05313 Stray constant definition detected at line 00003"
                          " of file .*empty_constant.insel"))

    def test_insel_include(self):
        self.assertEqual(insel.run('templates/insel_include.insel'), 3)

    def test_merging_two_loops(self):
        self.assertRaisesRegex(InselError, "Please try to merge 2 & 3", insel.run,
                               'templates/merge_distinct_loops.insel')

    def test_read_relative_file_when_in_correct_folder(self):
        with cwd(SCRIPT_DIR / 'templates'):
            deviation = insel.run('io/read_relative_file.insel')
            self.compareLists(deviation, [0, 0], places=4)

    def test_read_relative_file_when_in_another_folder(self):
        with cwd(SCRIPT_DIR):
            deviation = insel.run('templates/io/read_relative_file.insel')
            self.compareLists(deviation, [0, 0], places=4)

    def test_can_read_relative_file_with_absolute_path(self):
        with cwd(Path.home()):
            deviation = insel.run(
                SCRIPT_DIR / 'templates' / 'io' / 'read_relative_file.insel')
            self.compareLists(deviation, [0, 0], places=4)

    def test_string_parameter_in_vseit_should_not_be_cut(self):
        for model in ['short_string.vseit', 'long_string.vseit']:
            insel_model = insel.raw_run('-m', 'templates/io/' + model)
            string_params = [
                p for p in insel_model.split() if p.count("'") == 2]
            self.assertEqual(len(string_params), 2,
                             f"2 string parameters should be found in {model}")

    def test_screen_headline_should_be_displayed(self):
        for model in ['short_string.vseit', 'long_string.vseit']:
            out = insel.raw_run('templates/io/' + model)
            lines = out.splitlines()
            headline = next(line for line in lines if 'String' in line)
            self.assertTrue(len(headline) < 100,
                            f"Headline '{headline}' shouldn't be too long")

    def test_screen_utf8_header_should_be_displayed(self):
        out = insel.raw_run('templates/io/utf_headline.insel')
        self.assertTrue('T€st 12345' in out,
                        "Headline should be allowed to be in UTF-8")

    def test_mpp_without_top_of_loop(self):
        self.assertRaisesRegex(InselError, "No TOL-block", insel.run,
                               'templates/photovoltaic/mpp_without_top_of_loop.vseit')

    def test_mpp_with_top_of_loop(self):
        out = insel.raw_run('templates/photovoltaic/mpp_with_top_of_loop.vseit')
        self.assertRegex(out, r'Maximum at 17\.3 Volt and 52\.6 Watt')

    def test_algebraic_loop(self):
        self.assertRaisesRegex(InselError, "Algebraic loop detected", insel.run,
                               'templates/engine/sum_sum_do.insel')

    def test_algebraic_loop_with_do_do(self):
        self.skipTest("templates/engine/do_do.insel fails with SIGSEGV")

    def test_algebraic_loop_with_sum_sum(self):
        self.skipTest("templates/engine/sum_sum.insel fails with SIGSEGV")



import unittest
from insel import Insel
from .constants import IMPORTANT_BLOCKS


class TestInselDoc(unittest.TestCase):
    def test_insel_pdfs(self):
        doc_dir = Insel.dirname / 'doc'
        for basename in ['Tutorial', 'BlockReference', 'UserBlockReference',
                         'GettingStarted', 'ProgrammersGuide']:
            pdf = doc_dir / f"insel{basename}_en.pdf"
            self.assertTrue(pdf.exists(), f"{pdf} should exist")
            self.assertTrue(pdf.stat().st_size > 100_000,
                            f"{pdf} should be large enough")

    def test_insel_block_pdfs(self):
        doc_dir = Insel.dirname / 'doc' / 'inselBlocks'
        for basename in IMPORTANT_BLOCKS:
            pdf = doc_dir / f"{basename}.pdf"
            self.assertTrue(pdf.exists(), f"{pdf} should exist")
            self.assertTrue(pdf.stat().st_size > 10_000,
                            f"{pdf} should be large enough")



import tempfile
import re
import os
from calendar import monthrange
from pathlib import Path
import insel
from insel import Insel, InselError
from .custom_assertions import CustomAssertions
from .constants import SCRIPT_DIR, IS_WINDOWS

os.chdir(SCRIPT_DIR)


class TestBasicTemplates(CustomAssertions):
    def test_a_times_b(self):
        self.assertAlmostEqual(insel.template('a_times_b'), 9, places=6)
        # NOTE: .insel can be included in template_name, but doesn't have to.
        self.assertAlmostEqual(insel.template(
            'a_times_b.insel', a=4), 12, places=6)
        # NOTE: template path can also be absolute.
        self.assertAlmostEqual(insel.template(SCRIPT_DIR / 'templates' / 'a_times_b.insel',
                                              a=4, b=5),
                               20, places=6)

    def test_array_parameter(self):
        self.assertRaisesRegex(AttributeError, "UndefinedValue for 'x'", insel.template,
                               'array_parameters')
        self.assertRaisesRegex(AttributeError, "'x' should be an array.", insel.template,
                               'array_parameters', x=3)
        self.assertEqual(6, insel.template('array_parameters', x=[1, 2, 3]))
        self.assertEqual(6, insel.template(
            'array_parameters', x={0: 1, 1: 3, 2: 2}))


class TestTemplatesWithConstants(CustomAssertions):
    def test_vseit_is_a_template(self):
        # A Vseit model should be a template with default values
        self.assertEqual(3, insel.run('templates/constants/x_plus_y.vseit'))
        self.assertEqual(3, insel.template('constants/x_plus_y.vseit'))

    def test_setting_constants_in_vseit(self):
        self.assertEqual(5, insel.template(
            'constants/x_plus_y.vseit', x=2, y=3))
        self.assertEqual(4, insel.template('constants/x_plus_y.vseit', x=2))
        self.assertEqual(4, insel.template('constants/x_plus_y.vseit', y=3))

    def test_setting_string_constants_in_vseit(self):
        self.assertListEqual([72, 109, 173, 227, 257, 279, 301, 270, 200, 132, 83, 74],
                             insel.template('constants/string_constant'))
        self.assertListEqual([326, 300, 250, 168, 122, 101, 107, 141, 198, 244, 302, 335],
                             insel.template('constants/string_constant', location_name='Perth'))
        self.assertRaisesRegex(InselError, 'Location not found', insel.template,
                               'constants/string_constant', location_name='NotACity')

    def test_example_vseit(self):
        # PV in Nurnberg:
        self.compareLists([3866, 3652],
                          insel.template('constants/nurnberg.vseit'), places=0)
        # PV in Phoenix
        self.compareLists([6560, 6260],
                          insel.template('constants/nurnberg.vseit',
                                         Latitude=33,
                                         Longitude=-112,
                                         Tilt=10,
                                         Timezone=-7),
                          places=-1)

    def test_placeholder_over_constant(self):
        self.assertEqual(12, insel.template('constants/both',
                                            placeholder_x=5,
                                            placeholder_y=7))
        self.assertEqual(12, insel.template('constants/both', placeholder_x=5, placeholder_y=7,
                                            constant_x=0, constant_y=0))
        self.assertEqual(5, insel.template('constants/conflict', x=3, y=5))

    def test_both_placeholder_and_constant(self):
        self.assertEqual(3.7, insel.template('constants/same', x=1.2))


class TestTemplates(CustomAssertions):
    def test_empty_if(self):
        self.assertEqual(insel.template('empty_if'), [])

    def test_ifelsenan(self):
        odds_as_nans = insel.template('odds_as_nans')
        self.compareLists(odds_as_nans[::2], range(-10, 12, 2))
        for odd in odds_as_nans[1::2]:
            self.assertNaN(odd)

    def test_if(self):
        only_evens = insel.template('remove_odds')
        self.compareLists(only_evens, range(-10, 12, 2))

    def test_mtm_averages(self):
        # Those places used to have wrong averages
        places = ['Cambridge', 'Inuvik', 'Milagro',
                  'Naesgard', 'Nurnberg', 'Planes de Montecrist']
        for place in places:
            self.assertTrue(insel.template('weather/check_temperatures', location=place),
                            f'Wrong temperatures for {place}')

    def test_gengt_consistency(self):
        deviation = insel.template('weather/gengt_comparison')
        # NOTE: Depending on system, pseudo random values can vary very slightly.
        #       1e-4 really isn't any problem for °C or W/m²
        self.compareLists(deviation, [0, 0], places=4)

    def test_gengt_averages(self):
        irradiance_deviation, temperature_deviation = \
            insel.template('weather/gengt_monthly_averages')
        self.assertAlmostEqual(irradiance_deviation, 0, delta=5,
                               msg="Irradiance shouldnt vary by more than 5 W/m²")
        self.assertAlmostEqual(temperature_deviation, 0, delta=0.1,
                               msg="Temperature shouldnt vary by more than 0.1K")

    def test_aligned_screen_block(self):
        # Check that numbers displayed by SCREEN '*' are separated by at least one space
        # and that the decimal separators are aligned one above the other
        matrix = insel.template('expg')
        for i, row in zip(range(-20, 20), matrix):
            power = i / 2
            self.assertAlmostEqual(power, row[0], places=6)
            positive = 10 ** power
            negative = -10 ** (-power)
            self.assertAlmostEqual(positive, row[1], delta=positive / 1e6)
            self.assertAlmostEqual(negative, row[2], delta=-negative / 1e6)

        # Run again, this time checking the nicely formatted raw output.
        aligned = insel.raw_run('templates/expg.insel')
        output = [line for line in aligned.splitlines() if '    ' in line]
        self.assertTrue(len(output) > 30, 'Many lines should be returned.')

        dot_indices = set()
        columns = 5

        def indices(text, char):
            return [i for i, ltr in enumerate(text) if ltr == char]

        for line in output:
            dot_indices.update(indices(line, '.'))

        self.assertEqual(len(dot_indices), columns,
                         f"There should be {columns} nicely aligned decimal points")

    def test_updated_coordinates(self):
        v1_results = insel.template('photovoltaic/nurnberg_v1',
                                    latitude=49.5,
                                    old_longitude=-11.08,
                                    old_timezone=23
                                    )
        self.assertRegex(Insel.last_raw_output, '0 errors?, 4 warnings?')
        v2_results = insel.template('photovoltaic/nurnberg_v2',
                                    latitude=49.5,
                                    longitude=11.08,
                                    timezone=+1
                                    )
        self.compareLists(v1_results, [3865, 3645], places=-1)
        self.compareLists(v2_results, v1_results, places=2)

    def test_cumc(self):
        table = insel.template('conditional/sum')
        for month, (x, y) in zip(range(1, 13), table):
            _, days = monthrange(2021, month)
            self.assertAlmostEqual(month, x)
            self.assertAlmostEqual(days * (days + 1) / 2 * 24, y)

    def test_maxxc(self):
        table = insel.template('conditional/max')
        for month, (x, y) in zip(range(1, 13), table):
            _, days = monthrange(2021, month)
            self.assertAlmostEqual(month, x)
            self.assertAlmostEqual(days, y)

    def test_max_xy(self):
        table = insel.template('stats/max_xy')
        self.compareLists(table, [90, 1])
        self.assertEqual(insel.template('stats/max_xy_c'),
                         [[2020, 31], [2021, 31]])
        table = [v for r in insel.template('stats/max_xy_p') for v in r]
        self.compareLists(table, [90, 1, 360, 0], places=6)

    def test_min_xy(self):
        table = insel.template('stats/min_xy')
        self.compareLists(table, [270, -1])
        self.assertEqual(insel.template('stats/min_xy_c'),
                         [[2020, 29], [2021, 28]])
        table = [v for r in insel.template('stats/min_xy_p') for v in r]
        self.compareLists(table, [0, 0, 270, -1], places=6)

    def test_minnc(self):
        table = insel.template('conditional/min')
        for month, (x, y) in zip(range(1, 13), table):
            _, days = monthrange(2021, month)
            self.assertAlmostEqual(month, x)
            self.assertAlmostEqual(-days, y)

    def test_avec(self):
        table = insel.template('conditional/average')
        for month, (x, y) in zip(range(1, 13), table):
            _, days = monthrange(2021, month)
            self.assertAlmostEqual(month, x)
            self.assertAlmostEqual((days + 1) / 2, y)

    def test_parametric_average(self):
        self.assertEqual(1, insel.template('parametric/average'))

    def test_parametric_max(self):
        self.assertEqual(1, insel.template('parametric/max'))

    def test_parametric_min(self):
        self.assertEqual(1, insel.template('parametric/min'))

    def test_polyg(self):
        values = insel.template('mathematics/polyg')
        self.compareLists(values, [1, 1, 5, 4, -2, 0, 2, 4, 4, 4])

    def test_non_ascii_template(self):
        utf8_template = insel.Template('a_times_b_utf8', a=2, b=2)
        utf8_template.timeout = 5
        self.assertEqual(utf8_template.run(), 4)

        iso_template = insel.Template('a_times_b_iso8859', a=4, b=4)
        iso_template.timeout = 5
        self.assertEqual(iso_template.run(), 16)

    def test_sunpower_isc(self):
        self.assertRaisesRegex(AttributeError, "UndefinedValue", insel.template,
                               'photovoltaic/i_sc')  # Missing pv_id. STC by default
        self.assertIsNone(Insel.last_raw_output)
        spr_isc = insel.template('photovoltaic/i_sc', pv_id='008823')
        self.assertIsInstance(spr_isc, float)
        self.assertAlmostEqual(spr_isc, 5.87, places=2)

        self.assertAlmostEqual(insel.template('photovoltaic/i_sc', pv_id='003305'),
                               5.96, places=2)
        # TODO: More research is needed :)
        self.skipTest("""This spec fails, probably because of a too low
                'Temperature coeff of short-circuit current' in .bp files
                 .982E-7 in this example, instead of ~0.2E-3""")
        self.assertAlmostEqual(insel.template('photovoltaic/i_sc', pv_id='003305', temperature=70),
                               5.96 + (70 - 25) * 3.5e-3, places=2)

    def test_sunpower_uoc(self):
        # Missing pv_id. STC by default
        self.assertRaises(AttributeError, insel.template, 'photovoltaic/u_oc')
        self.assertAlmostEqual(insel.template('photovoltaic/u_oc', pv_id='003305'),
                               64.2, places=2)
        # TODO: More research is needed :)
        self.skipTest("Not sure about this calculation.")
        temp = 70
        self.assertAlmostEqual(insel.template('photovoltaic/u_oc', pv_id='003305',
                                              temperature=temp),
                               64.2 + (temp - 25) * (-0.1766), places=2)

    def test_sunpower_mpp(self):
        # Missing pv_id. STC by default
        self.assertRaises(AttributeError, insel.template, 'photovoltaic/mpp')
        self.assertAlmostEqual(insel.template('photovoltaic/mpp', pv_id='003305'),
                               305, places=0)
        temp = 70
        # TODO: Check with PVSYST or PVLIB. Is this the correct formula?
        # NOTE: -0.38%/K P_mpp, according to SPR 305 manual
        #       (https://www.pocosolar.com/wp-content/themes/twentyfifteen/pdfs/Sunpower%20Solar%20Panels/sunpower_305wht_spec_sheet.pdf)
        self.assertAlmostEqual(insel.template('photovoltaic/mpp', pv_id='003305', temperature=temp),
                               305 * (1 - 0.38 / 100) ** (temp - 25), places=0)

    def test_write_block(self):
        self._run_write_block()
        self._run_write_block(overwrite=0)
        self._run_write_block(overwrite=1)
        self._run_write_block(overwrite=2)

        self._run_write_block(basename='Ñüößç&txt.täxt€',
                              header='#ßeäöütµ§%&²³@°')

        self._run_write_block(basename='with a space.txt')
        self._run_write_block(basename='with_underscore.txt')

        self._run_write_block(fortran_format='(2F10.5)')

        self._run_write_block(header='#Some header here')

    def _run_write_block(self, basename='test.dat', **write_params):
        with tempfile.TemporaryDirectory() as tmpdirname:
            dat_file = Path(tmpdirname) / basename
            self.assertFalse(dat_file.exists())
            model = insel.Template(
                'io/write_params', dat_file=dat_file, **write_params)
            model.run()
            self.assertEqual(model.warnings, [])
            self.assertTrue(dat_file.exists(), "File should have been written")
            with open(dat_file) as out:
                if write_params.get('header'):
                    next(out)
                content = out.readlines()
                written = [float(line.split()[1]) for line in content]
                self.compareLists(
                    written, [x**2 for x in range(1, 11)], places=5)

    def test_write_block_relative_file_with_long_filename(self):
        """Template file is written in temp folder, in order for INSEL to run
        It means that relative files will be relative to temp folder"""
        dat_file = Path(tempfile.gettempdir(), 'a' * 50 + '.txt')
        dat_file.unlink(missing_ok=True)
        insel.template('io/write_params', dat_file=dat_file.name)
        self.assertTrue(dat_file.exists(),
                        f"{dat_file} should have been written")
        dat_file.unlink()

    def test_write_block_fails_if_folder_missing(self):
        filename = r'S:\\missing\\folder.txt' if IS_WINDOWS else '/not/here.txt'
        error_message = rf"(?m)^F05029 Block 00003: Cannot open file: {re.escape(filename)}$"

        self.assertRaisesRegex(InselError,
                               error_message,
                               insel.template, 'io/write_params',
                               dat_file=filename)

    def test_write_block_fails_if_read_only(self):
        if IS_WINDOWS:
            self.skipTest(
                "Docker tests for Windows are run as root, and ignore file permissions")
        with tempfile.TemporaryDirectory() as tmpdirname:
            dat_file = Path(tmpdirname) / 'read_only.dat'
            # Create read-only temp file:
            dat_file.touch()
            dat_file.chmod(mode=0o444)
            insel.template('io/write_params', dat_file=dat_file, overwrite=1)
            self.assertEqual(Insel.last_warnings, [
                'F05069 Block 00003: Unexpected write error - simulation terminated'])

    def test_read_simple_file(self):
        fourfivesix = insel.template('io/read_simple_file', ext='dat')
        self.compareLists(fourfivesix, [4, 5, 6])

    def test_read_missing_file(self):
        self.assertRaisesRegex(InselError, "(?m)^F05029 Block 00001: Cannot open file: .*not_here$",
                               insel.template, 'io/not_here')
        self.assertRegex(Insel.last_raw_output, '1 errors?, 0 warnings?')

    def test_read_too_many_lines(self):
        fourfivesix = insel.template('io/read_simple_file', ext='dat', lines=5)
        self.compareLists(fourfivesix, [4, 5, 6])
        self.assertEqual(Insel.last_warnings, [
                         'F05031 Block 00002: Unexpected end of file - simulation terminated'])

    def test_read_csv_like_as_normal_file(self):
        # READ block used to completely skip CSV files :-/
        # Now it just tries to read it as a normal file
        fourfivesix = insel.template('io/read_simple_file', ext='csv')
        self.compareLists(fourfivesix, [4, 5, 6])

    def test_read_epw_file(self):
        # Depending on extension, READ block will parse as normal file or EPW
        stuttgart_epw_average_temp = insel.template(
            'io/read_epw_file', ext='epw')
        self.assertAlmostEqual(stuttgart_epw_average_temp, 9.1, places=1)
        nothing_to_read = insel.template('io/read_epw_file', ext='txt')
        self.assertEqual(nothing_to_read, [])

    def test_weird_output(self):
        floats_and_lists = insel.template('io/weird_output')
        out = ';'.join(sorted(str(x) for x in floats_and_lists))
        self.assertEqual(
            out, '1.0;2.0;[1.0, 1.0, 1.0];[1.0, 1.0];[2.0, 2.0, 2.0];[2.0, 2.0]')



import subprocess
from .model import Model
from .insel import Insel


class ExistingModel(Model):
    def __init__(self, *params):
        super().__init__()
        self.params = list(params)

    def raw_results(self) -> bytes:
        Insel.calls += 1
        return subprocess.check_output([Insel.command] + self.params,
                                       shell=False)



from pathlib import Path
from typing import Dict, Any
import re
import tempfile
from .temporary_model import TemporaryModel
from .insel import Insel


class Template(TemporaryModel):
    """
    TODO: Document
    """
    # dirname is relative to current working directory.
    # NOTE: It should not be resolved yet, because CWD might change after "import insel"
    dirname: Path = Path('templates')
    # $ placeholder_example || 1.234 $
    placeholder_pattern = re.compile(r'\$([\w ]+)(?:\[(\d+)\] *)?(?:\|\|([\-\w \.\*]*))?\$')
    # C constant_example 1.234
    constants_pattern = re.compile(r'^C\s+(\w+)\s+(["\+\-\w \.\']+)(?:% .*)?\n', re.MULTILINE)
    # PLOT and PLOT2 gnuplot filenames
    gnuplot_pattern = re.compile(r'(?:advanced_plot|insel)\.gnu')
    # Will be used to disable gnuplot
    empty_gnuplot = Path(tempfile.gettempdir()) / 'empty.gnuplot'

    def __init__(self, template_path, **parameters) -> None:
        super().__init__()
        template_path = Path(template_path)
        if template_path.suffix == '.vseit':
            self.template_path: Path = template_path
        else:
            self.template_path: Path = template_path.with_suffix('.insel')
        self.name: str = self.template_path.stem
        self.parameters: Dict[str, Any] = self.add_defaults_to(parameters)

    def template_full_path(self) -> Path:
        full_path: Path = Template.dirname.resolve() / self.template_path
        if full_path.exists():
            return full_path
        else:
            raise FileNotFoundError("No template in %s" % full_path)

    def replace_placeholders(self, match_object: re.Match) -> str:
        var_name: str
        index: str
        default: str
        var_name, index, default = match_object.groups()
        var_name = var_name.strip()
        if var_name in self.parameters:
            if index:
                try:
                    return str(self.parameters[var_name][int(index)])
                except TypeError:
                    raise AttributeError(f"'{var_name}' should be an array.")
            else:
                return str(self.parameters[var_name])
        elif default is not None:
            return default
        else:
            raise AttributeError(
                "UndefinedValue for '%s' in %s.insel template" %
                (var_name, self.name))

    def replace_constants(self, match_object: re.Match) -> str:
        var_name: str
        default: str
        var_name, default = match_object.groups()
        var_name = var_name.strip()
        if var_name in self.parameters:
            value =  repr(self.parameters[var_name])
        else:
            value = default
        return f"C {var_name} {value}"

    def disable_gnuplot(self, content):
        content, count = re.subn(Template.gnuplot_pattern, Template.empty_gnuplot.as_posix(), content)
        if count:
            with open(Template.empty_gnuplot, 'w') as tmp_gnuplot:
                tmp_gnuplot.write("exit\n")
        return content

    def add_defaults_to(self, parameters):
        defaults = {
            'bp_folder': Insel.dirname / "data" / "bp",
            'data_folder': Template.dirname.parent / "data",
            'template_folder': Template.dirname,
            'gnuplot': False
        }
        defaults.update(parameters)
        return defaults

    def content(self) -> str:
        # Replace unknown chars with backslash + code, so that content can be fed to INSEL
        with open(self.template_full_path(),
                  encoding='utf-8',
                  errors='backslashreplace') as template:
            content = template.read()
            content = re.sub(Template.constants_pattern, self.replace_constants, content)
            content = re.sub(Template.placeholder_pattern, self.replace_placeholders, content)
            if not self.parameters['gnuplot']:
                content = self.disable_gnuplot(content)
            #NOTE: Test if variable hasn't been used?
            return content



# coding=utf8
import insel
from .custom_assertions import CustomAssertions


class TestUserBlocks(CustomAssertions):
    def test_ubstorage(self):
        insel.block('ubstorage', 1, 2, parameters=[
                    10, 0, 1, 1, 0, 100, 0, 1, 0])

    def test_ubisonland(self):
        self.assertAlmostEqual(insel.block('ubisonland', 48.77, 9.18), 1)
        self.assertAlmostEqual(insel.block('ubisonland', 48.77, -9.18), 0)

    # TODO: Test UBCHP
